/**
 @file bzrtpCryptoTests.h

 @author Johan Pascal

 @copyright Copyright (C) 2014 Belledonne Communications, Grenoble, France
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
void test_RNG(void);
void test_sha256(void);
void test_hmacSha256(void);
void test_aes128CFB(void);
void test_dhm2048(void);
void test_dhm3072(void);
void test_zrtpKDF(void);
void test_CRC32(void);
void test_algoAgreement(void);
void test_algoSetterGetter(void);
void test_addMandatoryCryptoTypesIfNeeded(void);
